import 'dart:async';
import 'package:oobium_datastore/oobium_datastore.dart';
import 'package:oobium_logging/oobium_logging.dart';
part 'models_test.schema.g.feature_persist.dart';
part 'models_test.schema.g.feature_rest.dart';
part 'models_test.schema.x.dart';

final _log = Logger(ModelsTestData);

class ModelsTestData extends ModelContext {
  ModelsTestData(
      {String? path,
      OnUpgrade? onUpgrade,
      SettingsRestAdapter? settingsRestAdapter,
      ConfigRestAdapter? configRestAdapter,
      AccountRestAdapter? accountRestAdapter,
      ImageRestAdapter? imageRestAdapter,
      UserRestAdapter? userRestAdapter,
      PostRestAdapter? postRestAdapter,
      CommentRestAdapter? commentRestAdapter})
      : super(
            StateContext(features: [
              _RestFeature(
                settingsRestAdapter:
                    settingsRestAdapter ?? SettingsRestAdapter(),
                configRestAdapter: configRestAdapter ?? ConfigRestAdapter(),
                accountRestAdapter: accountRestAdapter ?? AccountRestAdapter(),
                imageRestAdapter: imageRestAdapter ?? ImageRestAdapter(),
                userRestAdapter: userRestAdapter ?? UserRestAdapter(),
                postRestAdapter: postRestAdapter ?? PostRestAdapter(),
                commentRestAdapter: commentRestAdapter ?? CommentRestAdapter(),
              ),
              _ModelFeature(),
              _TriggerFeature(),
              if (path != null)
                _PersistFeature(
                    path: path,
                    version: 1,
                    compactionStrategy: const DefaultCompactionStrategy(),
                    converter: const _Converter(),
                    onUpgrade: onUpgrade ?? dropOnUpgrade)
            ]),
            const _Converter());

  Future<ModelsTestData> open() => super.open().then((_) => this);

  SettingsRef get settings => feature<_ModelFeature>().settings;
  ConfigRef get config => feature<_ModelFeature>().config;
  Accounts get accounts => feature<_ModelFeature>().accounts;
  Images get images => feature<_ModelFeature>().images;
  Users get users => feature<_ModelFeature>().users;
  Posts get posts => feature<_ModelFeature>().posts;
  Comments get comments => feature<_ModelFeature>().comments;

  void load(data) => feature<_ModelFeature>().load(data);
  void unload(data) => feature<_ModelFeature>().unload(data);
}

class _Converter extends Converter {
  const _Converter();

  @override
  StateValue decodePut(DataRecord r) {
    final data = r.toJsonData();
    switch (r.type) {
      case 'Settings':
        return Settings._(
            data['name'],
            ((data['account'] != null)
                ? AccountKey.value(data['account'])
                : null),
            ((data['user'] != null) ? data['user'] : null));
      case 'Config':
        return Config._(
            data['name'],
            (data['options'] ?? {})
                .map((k, v) => MapEntry('$k', v))
                .cast<String, String, dynamic>());
      case 'Account':
        return Account._(AccountKey.value(r.modelId), data['name'],
            ((data['settings'] != null) ? data['settings'] : null));
      case 'Image':
        return Image._(
            ImageKey.value(r.modelId),
            ((data['name'] != null) ? data['name'] : null),
            data['height'],
            data['width']);
      case 'User':
        return User._(
            UserKey.value(r.modelId),
            data['name'],
            ((data['desc'] != null) ? data['desc'] : null),
            Address.fromJson(data['home']),
            ((data['work'] != null) ? Address.fromJson(data['work']) : null));
      case 'Post':
        return Post._(PostKey.value(r.modelId), data['title'], data['user']);
      case 'Comment':
        return Comment._(
            CommentKey.value(r.modelId),
            data['post'],
            data['message'],
            data['owner'],
            ((data['auditor'] != null) ? data['auditor'] : null));
      default:
        throw Exception('unknown type ${r.type} during decodePut');
    }
  }

  @override
  StateKey decodeRemove(DataRecord r) {
    switch (r.type) {
      case 'Settings':
        return SettingsKey();
      case 'Config':
        return ConfigKey();
      case 'Account':
        return AccountKey.parse(r.modelId);
      case 'Image':
        return ImageKey.parse(r.modelId);
      case 'User':
        return UserKey.parse(r.modelId);
      case 'Post':
        return PostKey.parse(r.modelId);
      case 'Comment':
        return CommentKey.parse(r.modelId);
      default:
        throw Exception('unknown type ${r.type} during decodeRemove');
    }
  }

  @override
  DataRecord encodePut(StateValue v) {
    if (v is Settings) {
      return DataRecord('Settings', 'Settings', '', {
        'name': v.name,
        if (v.account != null) 'account': v.account?.hexString,
        if (v.user != null) 'user': v.user
      });
    } else if (v is Config) {
      return DataRecord('Config', 'Config', '',
          {'name': v.name, 'options': recordValue(v.options)});
    } else if (v is Account) {
      return DataRecord('Account', '${v.id}', '',
          {'name': v.name, if (v.settings != null) 'settings': v.settings});
    } else if (v is Image) {
      return DataRecord('Image', '${v.id}', '', {
        if (v.name != null) 'name': v.name,
        'height': v.height,
        'width': v.width
      });
    } else if (v is User) {
      return DataRecord('User', '${v.id}', '', {
        'name': v.name,
        if (v.desc != null) 'desc': v.desc,
        'home': v.home.toJson(),
        if (v.work != null) 'work': v.work?.toJson()
      });
    } else if (v is Post) {
      return DataRecord(
          'Post', '${v.id}', '', {'title': v.title, 'user': v.user});
    } else if (v is Comment) {
      return DataRecord('Comment', '${v.id}', '', {
        'post': v.post,
        'message': v.message,
        'owner': v.owner,
        if (v.auditor != null) 'auditor': v.auditor
      });
    } else {
      throw Exception('unknown type $v during encodePut');
    }
  }

  @override
  DataRecord encodeRemove(StateValue k) {
    if (k is Settings) {
      return DataRecord('Settings', 'Settings', '');
    } else if (k is Config) {
      return DataRecord('Config', 'Config', '');
    } else if (k is Account) {
      return DataRecord('Account', '${k.id}', '');
    } else if (k is Image) {
      return DataRecord('Image', '${k.id}', '');
    } else if (k is User) {
      return DataRecord('User', '${k.id}', '');
    } else if (k is Post) {
      return DataRecord('Post', '${k.id}', '');
    } else if (k is Comment) {
      return DataRecord('Comment', '${k.id}', '');
    } else {
      throw Exception('unknown type $k during encodeRemove');
    }
  }
}

class Address implements Comparable<Address> {
  Address(
      {required String street1,
      String? street2,
      required String city,
      required String state,
      required String zipcode,
      Map<String, dynamic> extras = const {}})
      : this._(street1, street2, city, state, zipcode, extras);
  Address.fromJson(Map<String, dynamic> data,
      {String? street1,
      String? street2,
      bool street2AsNull = false,
      String? city,
      String? state,
      String? zipcode,
      Map<String, dynamic>? extras})
      : this._(
            street1 ?? data['street1'],
            street2 ?? ((data['street2'] != null) ? data['street2'] : null),
            city ?? data['city'],
            state ?? data['state'],
            zipcode ?? data['zipcode'],
            extras ??
                ((data['extras'] != null)
                    ? (data['extras'] ?? {})
                        .map((k, v) => MapEntry('$k', v))
                        .cast<String, String, dynamic>()
                    : const {}));
  static List<Address> fromJsonList(list,
          {String? street1,
          String? street2,
          bool street2AsNull = false,
          String? city,
          String? state,
          String? zipcode,
          Map<String, dynamic>? extras}) =>
      (list as Iterable)
          .map((data) => Address.fromJson(data,
              street1: street1,
              street2: street2,
              street2AsNull: street2AsNull,
              city: city,
              state: state,
              zipcode: zipcode,
              extras: extras))
          .toList();
  Address._(this.street1, this.street2, this.city, this.state, this.zipcode,
      this.extras);

  final String street1;
  final String? street2;
  final String city;
  final String state;
  final String zipcode;
  final Map<String, dynamic> extras;

  @override
  int compareTo(Address other) =>
      compare(street1, other.street1) ??
      compare(street2, other.street2) ??
      compare(city, other.city) ??
      compare(state, other.state) ??
      compare(zipcode, other.zipcode) ??
      compare(extras, other.extras) ??
      0;

  Address copyWith(
          {String? street1,
          String? street2,
          bool street2AsNull = false,
          String? city,
          String? state,
          String? zipcode,
          Map<String, dynamic>? extras}) =>
      Address._(
          street1 ?? this.street1,
          street2AsNull ? null : (street2 ?? this.street2),
          city ?? this.city,
          state ?? this.state,
          zipcode ?? this.zipcode,
          extras ?? this.extras);

  @override
  bool operator ==(Object? other) =>
      identical(this, other) ||
      (other is Address &&
          (street1 == other.street1 ||
              street2 == other.street2 ||
              city == other.city ||
              state == other.state ||
              zipcode == other.zipcode ||
              extras == other.extras));

  @override
  int get hashCode =>
      Object.hash(Address, street1, street2, city, state, zipcode, extras);

  Map<String, dynamic> toJson() => {
        'street1': street1,
        if (street2 != null) 'street2': street2,
        'city': city,
        'state': state,
        'zipcode': zipcode,
        'extras': jsonValue(extras)
      };
}

class SettingsKey extends StateKey<Settings> {
  @override
  final String id = 'Settings';
  const SettingsKey();
}

class Settings extends Singleton<Settings> implements SettingsKey {
  Settings(
      {String name = 'initial settings name', ObjectId? account, int? user})
      : this._(name, account, user);
  Settings.fromJson(Map<String, dynamic> data,
      {String? name,
      ObjectId? account,
      bool accountAsNull = false,
      int? user,
      bool userAsNull = false})
      : this._(
            name ??
                ((data['name'] != null)
                    ? data['name']
                    : 'initial settings name'),
            account ??
                ((data['account'] != null)
                    ? AccountKey.value(data['account'])
                    : null),
            user ?? ((data['user'] != null) ? data['user'] : null));
  static List<Settings> fromJsonList(list,
          {String? name,
          ObjectId? account,
          bool accountAsNull = false,
          int? user,
          bool userAsNull = false}) =>
      (list as Iterable)
          .map((data) => Settings.fromJson(data,
              name: name,
              account: account,
              accountAsNull: accountAsNull,
              user: user,
              userAsNull: userAsNull))
          .toList();
  Settings._(this.name, this.account, this.user);

  final String name;
  final ObjectId? account;
  final int? user;

  @override
  int compareTo(Settings other) =>
      compare(name, other.name) ??
      compare(account, other.account) ??
      compare(user, other.user) ??
      0;

  Settings copyWith(
          {String? name,
          ObjectId? account,
          bool accountAsNull = false,
          int? user,
          bool userAsNull = false}) =>
      Settings._(
          name ?? this.name,
          accountAsNull ? null : (account ?? this.account),
          userAsNull ? null : (user ?? this.user));

  @override
  Map<String, dynamic> toJson() => {
        'id': id,
        'name': name,
        if (account != null) 'account': account?.hexString,
        if (user != null) 'user': user
      };
}

class SettingsRef extends SingletonRef<Settings, SettingsRef>
    implements SettingsKey {
  final _ModelFeature _feature;
  SettingsRef._(this._feature) : super(_feature);

  String get name => _feature.get(this).name;
  set name(String value) => update(name: value);
  AccountRef? get account {
    final v = _feature.get(this).account;
    return (v != null) ? AccountRef._(_feature, v) : null;
  }

  set account(AccountRef? value) =>
      (value != null) ? update(account: value) : update(accountAsNull: true);
  UserRef? get user {
    final v = _feature.get(this).user;
    return (v != null) ? UserRef._(_feature, v) : null;
  }

  set user(UserRef? value) =>
      (value != null) ? update(user: value) : update(userAsNull: true);

  Future<SettingsRef> fetch({bool force = false}) => _feature<_RestFeature>()
      .settingsRestAdapter
      .fetch(force)
      .then((_) => this);

  void update(
      {String? name,
      AccountKey? account,
      bool accountAsNull = false,
      UserKey? user,
      bool userAsNull = false}) {
    final s = Settings._(
        (name ?? this.name),
        accountAsNull ? null : (account ?? this.account)?.id,
        userAsNull ? null : (user ?? this.user)?.id);
    _feature.put(s);
  }
}

class ConfigKey extends StateKey<Config> {
  @override
  final String id = 'Config';
  const ConfigKey();
}

class Config extends Singleton<Config> implements ConfigKey {
  Config({required String name, Map<String, dynamic> options = const {}})
      : this._(name, options);
  Config.fromJson(Map<String, dynamic> data,
      {String? name, Map<String, dynamic>? options})
      : this._(
            name ?? data['name'],
            options ??
                ((data['options'] != null)
                    ? (data['options'] ?? {})
                        .map((k, v) => MapEntry('$k', v))
                        .cast<String, String, dynamic>()
                    : const {}));
  static List<Config> fromJsonList(list,
          {String? name, Map<String, dynamic>? options}) =>
      (list as Iterable)
          .map((data) => Config.fromJson(data, name: name, options: options))
          .toList();
  Config._(this.name, this.options);

  final String name;
  final Map<String, dynamic> options;

  @override
  int compareTo(Config other) =>
      compare(name, other.name) ?? compare(options, other.options) ?? 0;

  Config copyWith({String? name, Map<String, dynamic>? options}) =>
      Config._(name ?? this.name, options ?? this.options);

  @override
  Map<String, dynamic> toJson() =>
      {'id': id, 'name': name, 'options': jsonValue(options)};
}

class ConfigRef extends SingletonRef<Config, ConfigRef> implements ConfigKey {
  final _ModelFeature _feature;
  ConfigRef._(this._feature) : super(_feature);

  String get name => _feature.get(this).name;
  set name(String value) => update(name: value);
  Map<String, dynamic> get options => _feature.get(this).options;
  set options(Map<String, dynamic> value) => update(options: value);

  Future<ConfigRef> fetch({bool force = false}) =>
      _feature<_RestFeature>().configRestAdapter.fetch(force).then((_) => this);

  void update({String? name, Map<String, dynamic>? options}) {
    final s = Config._((name ?? this.name), (options ?? this.options));
    _feature.put(s);
  }
}

class AccountKey extends StateKey<Account> {
  @override
  final ObjectId id;
  AccountKey([ObjectId? id]) : id = id ?? ObjectId();
  AccountKey.parse(String s) : id = AccountKey.value(s);
  static ObjectId value(String s) => ObjectId.fromHexString(s);
  static ObjectId? maybe(String? s) {
    try {
      return (s != null) ? AccountKey.value(s) : null;
    } catch (_) {
      return null;
    }
  }
}

class Account extends Model<Account, ObjectId> implements AccountKey {
  Account({ObjectId? id, required String name, String? settings})
      : this._(id ?? ObjectId(), name, settings);
  Account.fromJson(Map<String, dynamic> data,
      {ObjectId? id,
      String? name,
      String? settings,
      bool settingsAsNull = false,
      AccountQuery? withQuery})
      : this._(
            id ??
                ((data['id'] != null)
                    ? AccountKey.value(data['id'])
                    : ObjectId()),
            name ?? withQuery?.name ?? data['name'],
            settings ??
                withQuery?.settings ??
                ((data['settings'] != null) ? data['settings'] : null));
  static List<Account> fromJsonList(list,
          {ObjectId Function(Map<String, dynamic> item)? id,
          String? name,
          String? settings,
          bool settingsAsNull = false,
          AccountQuery? withQuery}) =>
      (list as Iterable)
          .map((data) => Account.fromJson(data,
              id: id?.call(data),
              name: name,
              settings: settings,
              settingsAsNull: settingsAsNull,
              withQuery: withQuery))
          .toList();
  Account._(ObjectId id, this.name, this.settings) : super(id);

  final String name;
  final String? settings;

  @override
  int compareTo(Account other) =>
      compare(name, other.name) ?? compare(settings, other.settings) ?? 0;

  @override
  Account copyWith(
          {ObjectId? id,
          String? name,
          String? settings,
          bool settingsAsNull = false}) =>
      Account._(id ?? this.id, name ?? this.name,
          settingsAsNull ? null : (settings ?? this.settings));

  @override
  Map<String, dynamic> toJson() =>
      {'id': id, 'name': name, if (settings != null) 'settings': settings};
}

class AccountRef extends ModelRef<Account, ObjectId, AccountRef>
    implements AccountKey {
  final _ModelFeature _feature;
  AccountRef._(this._feature, ObjectId id) : super(_feature, id);

  String get name => _feature.get(this).name;
  set name(String value) => update(name: value);
  SettingsRef? get settings =>
      (_feature.get(this).settings != null) ? SettingsRef._(_feature) : null;
  set settings(SettingsRef? value) =>
      (value != null) ? update(settings: value) : update(settingsAsNull: true);
  void update(
      {String? name, SettingsKey? settings, bool settingsAsNull = false}) {
    final s = Account._(id, (name ?? this.name),
        settingsAsNull ? null : (settings ?? this.settings)?.id);
    _feature.put(s);
  }
}

class AccountQuery {
  final ObjectId? id;
  final String? name;
  final String? settings;
  final bool? settingsIsNull;

  AccountQuery({this.id, this.name, this.settingsIsNull, this.settings});
  AccountQuery._(this.id, this.name, this.settingsIsNull, this.settings);

  bool get isAll =>
      (id == null) &&
      (name == null) &&
      (settingsIsNull == null) &&
      (settings == null);
  bool get isNotAll => !isAll;

  bool test(Account m) =>
      (id == null || id == m.id) &&
      (name == null || name == m.name) &&
      ((settingsIsNull == true)
          ? (m.settings == null)
          : (settingsIsNull == false)
              ? (settings == null)
                  ? (m.settings != null)
                  : (settings == m.settings)
              : ((settings == null) || (settings == m.settings)));

  @override
  bool operator ==(Object? other) =>
      identical(this, other) ||
      (other is Account &&
          id == other.id &&
          name == other.name &&
          settings == other.settings);

  @override
  int get hashCode => Object.hash(Account, id, name, settings);
}

class ImageKey extends StateKey<Image> {
  @override
  final int id;
  const ImageKey(this.id);
  ImageKey.parse(String s) : id = ImageKey.value(s);
  static int value(String s) => int.parse(s);
  static int? maybe(String? s) {
    try {
      return (s != null) ? ImageKey.value(s) : null;
    } catch (_) {
      return null;
    }
  }
}

class Image extends Model<Image, int> implements ImageKey {
  Image(
      {required int id, String? name, required int height, required int width})
      : this._(id, name, height, width);
  Image.fromJson(Map<String, dynamic> data,
      {int? id,
      String? name,
      bool nameAsNull = false,
      int? height,
      int? width,
      ImageQuery? withQuery})
      : this._(
            id ?? (data['id'] as int),
            name ??
                withQuery?.name ??
                ((data['name'] != null) ? data['name'] : null),
            height ?? withQuery?.height ?? data['height'],
            width ?? withQuery?.width ?? data['width']);
  static List<Image> fromJsonList(list,
          {int Function(Map<String, dynamic> item)? id,
          String? name,
          bool nameAsNull = false,
          int? height,
          int? width,
          ImageQuery? withQuery}) =>
      (list as Iterable)
          .map((data) => Image.fromJson(data,
              id: id?.call(data),
              name: name,
              nameAsNull: nameAsNull,
              height: height,
              width: width,
              withQuery: withQuery))
          .toList();
  Image._(int id, this.name, this.height, this.width) : super(id);

  final String? name;
  final int height;
  final int width;

  @override
  int compareTo(Image other) =>
      compare(name, other.name) ??
      compare(height, other.height) ??
      compare(width, other.width) ??
      0;

  @override
  Image copyWith(
          {int? id,
          String? name,
          bool nameAsNull = false,
          int? height,
          int? width}) =>
      Image._(id ?? this.id, nameAsNull ? null : (name ?? this.name),
          height ?? this.height, width ?? this.width);

  @override
  Map<String, dynamic> toJson() => {
        'id': id,
        if (name != null) 'name': name,
        'height': height,
        'width': width
      };
}

class ImageRef extends ModelRef<Image, int, ImageRef> implements ImageKey {
  final _ModelFeature _feature;
  ImageRef._(this._feature, int id) : super(_feature, id);

  String? get name => _feature.get(this).name;
  set name(String? value) =>
      (value != null) ? update(name: value) : update(nameAsNull: true);
  int get height => _feature.get(this).height;
  set height(int value) => update(height: value);
  int get width => _feature.get(this).width;
  set width(int value) => update(width: value);

  Future<ImageRef> fetch({bool force = false}) => _feature<_RestFeature>()
      .imageRestAdapter
      .fetch(ImageQuery(id: id), force)
      .then((_) => this);

  void update(
      {String? name, bool nameAsNull = false, int? height, int? width}) {
    final s = Image._(id, nameAsNull ? null : (name ?? this.name),
        (height ?? this.height), (width ?? this.width));
    _feature.put(s);
  }
}

class ImageQuery {
  final int? id;
  final String? name;
  final bool? nameIsNull;
  final int? height;
  final int? width;

  ImageQuery({this.id, this.nameIsNull, this.name, this.height, this.width});
  ImageQuery._(this.id, this.nameIsNull, this.name, this.height, this.width);

  bool get isAll =>
      (id == null) &&
      (nameIsNull == null) &&
      (name == null) &&
      (height == null) &&
      (width == null);
  bool get isNotAll => !isAll;

  bool test(Image m) =>
      (id == null || id == m.id) &&
      ((nameIsNull == true)
          ? (m.name == null)
          : (nameIsNull == false)
              ? (name == null)
                  ? (m.name != null)
                  : (name == m.name)
              : ((name == null) || (name == m.name))) &&
      (height == null || height == m.height) &&
      (width == null || width == m.width);

  @override
  bool operator ==(Object? other) =>
      identical(this, other) ||
      (other is Image &&
          id == other.id &&
          name == other.name &&
          height == other.height &&
          width == other.width);

  @override
  int get hashCode => Object.hash(Image, id, name, height, width);
}

class UserKey extends StateKey<User> {
  @override
  final int id;
  const UserKey(this.id);
  UserKey.parse(String s) : id = UserKey.value(s);
  static int value(String s) => int.parse(s);
  static int? maybe(String? s) {
    try {
      return (s != null) ? UserKey.value(s) : null;
    } catch (_) {
      return null;
    }
  }
}

class User extends Model<User, int> implements UserKey {
  User(
      {required int id,
      required String name,
      String? desc,
      required Address home,
      Address? work})
      : this._(id, name, desc, home, work);
  User.fromJson(Map<String, dynamic> data,
      {int? id,
      String? name,
      String? desc,
      bool descAsNull = false,
      Address? home,
      Address? work,
      bool workAsNull = false,
      UserQuery? withQuery})
      : this._(
            id ?? (data['id'] as int),
            name ?? withQuery?.name ?? data['name'],
            desc ??
                withQuery?.desc ??
                ((data['desc'] != null) ? data['desc'] : null),
            home ?? withQuery?.home ?? Address.fromJson(data['home']),
            work ??
                withQuery?.work ??
                ((data['work'] != null)
                    ? Address.fromJson(data['work'])
                    : null));
  static List<User> fromJsonList(list,
          {int Function(Map<String, dynamic> item)? id,
          String? name,
          String? desc,
          bool descAsNull = false,
          Address? home,
          Address? work,
          bool workAsNull = false,
          UserQuery? withQuery}) =>
      (list as Iterable)
          .map((data) => User.fromJson(data,
              id: id?.call(data),
              name: name,
              desc: desc,
              descAsNull: descAsNull,
              home: home,
              work: work,
              workAsNull: workAsNull,
              withQuery: withQuery))
          .toList();
  User._(int id, this.name, this.desc, this.home, this.work) : super(id);

  final String name;
  final String? desc;
  final Address home;
  final Address? work;

  @override
  int compareTo(User other) =>
      compare(name, other.name) ??
      compare(desc, other.desc) ??
      compare(home, other.home) ??
      compare(work, other.work) ??
      0;

  @override
  User copyWith(
          {int? id,
          String? name,
          String? desc,
          bool descAsNull = false,
          Address? home,
          Address? work,
          bool workAsNull = false}) =>
      User._(
          id ?? this.id,
          name ?? this.name,
          descAsNull ? null : (desc ?? this.desc),
          home ?? this.home,
          workAsNull ? null : (work ?? this.work));

  @override
  Map<String, dynamic> toJson() => {
        'id': id,
        'name': name,
        if (desc != null) 'desc': desc,
        'home': home.toJson(),
        if (work != null) 'work': work?.toJson()
      };
}

class UserRef extends ModelRef<User, int, UserRef> implements UserKey {
  final _ModelFeature _feature;
  UserRef._(this._feature, int id) : super(_feature, id);

  String get name => _feature.get(this).name;
  set name(String value) => update(name: value);
  String? get desc => _feature.get(this).desc;
  set desc(String? value) =>
      (value != null) ? update(desc: value) : update(descAsNull: true);
  Address get home => _feature.get(this).home;
  set home(Address value) => update(home: value);
  Address? get work => _feature.get(this).work;
  set work(Address? value) =>
      (value != null) ? update(work: value) : update(workAsNull: true);
  UserPosts get posts => UserPosts(_feature, this);

  Future<UserRef> fetch({bool force = false}) => _feature<_RestFeature>()
      .userRestAdapter
      .fetch(UserQuery(id: id), force)
      .then((_) => this);

  void update(
      {String? name,
      String? desc,
      bool descAsNull = false,
      Address? home,
      Address? work,
      bool workAsNull = false}) {
    final s = User._(
        id,
        (name ?? this.name),
        descAsNull ? null : (desc ?? this.desc),
        (home ?? this.home),
        workAsNull ? null : (work ?? this.work));
    _feature.put(s);
  }
}

class UserQuery {
  final int? id;
  final String? name;
  final String? desc;
  final bool? descIsNull;
  final Address? home;
  final Address? work;
  final bool? workIsNull;

  UserQuery(
      {this.id,
      this.name,
      this.descIsNull,
      this.desc,
      this.home,
      this.workIsNull,
      this.work});
  UserQuery._(this.id, this.name, this.descIsNull, this.desc, this.home,
      this.workIsNull, this.work);

  bool get isAll =>
      (id == null) &&
      (name == null) &&
      (descIsNull == null) &&
      (desc == null) &&
      (home == null) &&
      (workIsNull == null) &&
      (work == null);
  bool get isNotAll => !isAll;

  bool test(User m) =>
      (id == null || id == m.id) &&
      (name == null || name == m.name) &&
      ((descIsNull == true)
          ? (m.desc == null)
          : (descIsNull == false)
              ? (desc == null)
                  ? (m.desc != null)
                  : (desc == m.desc)
              : ((desc == null) || (desc == m.desc))) &&
      (home == null || home == m.home) &&
      ((workIsNull == true)
          ? (m.work == null)
          : (workIsNull == false)
              ? (work == null)
                  ? (m.work != null)
                  : (work == m.work)
              : ((work == null) || (work == m.work)));

  @override
  bool operator ==(Object? other) =>
      identical(this, other) ||
      (other is User &&
          id == other.id &&
          name == other.name &&
          desc == other.desc &&
          home == other.home &&
          work == other.work);

  @override
  int get hashCode => Object.hash(User, id, name, desc, home, work);
}

class PostKey extends StateKey<Post> {
  @override
  final int id;
  const PostKey(this.id);
  PostKey.parse(String s) : id = PostKey.value(s);
  static int value(String s) => int.parse(s);
  static int? maybe(String? s) {
    try {
      return (s != null) ? PostKey.value(s) : null;
    } catch (_) {
      return null;
    }
  }
}

class Post extends Model<Post, int> implements PostKey {
  Post({required int id, required String title, required int user})
      : this._(id, title, user);
  Post.fromJson(Map<String, dynamic> data,
      {int? id, String? title, int? user, PostQuery? withQuery})
      : this._(
            id ?? (data['id'] as int),
            title ?? withQuery?.title ?? data['title'],
            user ?? withQuery?.user ?? data['user']);
  static List<Post> fromJsonList(list,
          {int Function(Map<String, dynamic> item)? id,
          String? title,
          int? user,
          PostQuery? withQuery}) =>
      (list as Iterable)
          .map((data) => Post.fromJson(data,
              id: id?.call(data),
              title: title,
              user: user,
              withQuery: withQuery))
          .toList();
  Post._(int id, this.title, this.user) : super(id);

  final String title;
  final int user;

  @override
  int compareTo(Post other) =>
      compare(title, other.title) ?? compare(user, other.user) ?? 0;

  @override
  Post copyWith({int? id, String? title, int? user}) =>
      Post._(id ?? this.id, title ?? this.title, user ?? this.user);

  @override
  Map<String, dynamic> toJson() => {'id': id, 'title': title, 'user': user};
}

class PostRef extends ModelRef<Post, int, PostRef> implements PostKey {
  final _ModelFeature _feature;
  PostRef._(this._feature, int id) : super(_feature, id);

  String get title => _feature.get(this).title;
  set title(String value) => update(title: value);
  UserRef get user => UserRef._(_feature, _feature.get(this).user);
  set user(UserRef value) => update(user: value);
  PostComments get comments => PostComments(_feature, this);

  Future<PostRef> fetch({bool force = false}) => _feature<_RestFeature>()
      .postRestAdapter
      .fetch(PostQuery(id: id), force)
      .then((_) => this);

  void update({String? title, UserKey? user}) {
    final s = Post._(id, (title ?? this.title), (user ?? this.user).id);
    _feature.put(s);
  }
}

class PostQuery {
  final int? id;
  final String? title;
  final int? user;

  PostQuery({this.id, this.title, this.user});
  PostQuery._(this.id, this.title, this.user);

  bool get isAll => (id == null) && (title == null) && (user == null);
  bool get isNotAll => !isAll;

  bool test(Post m) =>
      (id == null || id == m.id) &&
      (title == null || title == m.title) &&
      (user == null || user == m.user);

  @override
  bool operator ==(Object? other) =>
      identical(this, other) ||
      (other is Post &&
          id == other.id &&
          title == other.title &&
          user == other.user);

  @override
  int get hashCode => Object.hash(Post, id, title, user);
}

class CommentKey extends StateKey<Comment> {
  @override
  final int id;
  const CommentKey(this.id);
  CommentKey.parse(String s) : id = CommentKey.value(s);
  static int value(String s) => int.parse(s);
  static int? maybe(String? s) {
    try {
      return (s != null) ? CommentKey.value(s) : null;
    } catch (_) {
      return null;
    }
  }
}

class Comment extends Model<Comment, int> implements CommentKey {
  Comment(
      {required int id,
      required int post,
      required String message,
      required int owner,
      int? auditor})
      : this._(id, post, message, owner, auditor);
  Comment.fromJson(Map<String, dynamic> data,
      {int? id,
      int? post,
      String? message,
      int? owner,
      int? auditor,
      bool auditorAsNull = false,
      CommentQuery? withQuery})
      : this._(
            id ?? (data['id'] as int),
            post ?? withQuery?.post ?? data['post'],
            message ?? withQuery?.message ?? data['message'],
            owner ?? withQuery?.owner ?? data['owner'],
            auditor ??
                withQuery?.auditor ??
                ((data['auditor'] != null) ? data['auditor'] : null));
  static List<Comment> fromJsonList(list,
          {int Function(Map<String, dynamic> item)? id,
          int? post,
          String? message,
          int? owner,
          int? auditor,
          bool auditorAsNull = false,
          CommentQuery? withQuery}) =>
      (list as Iterable)
          .map((data) => Comment.fromJson(data,
              id: id?.call(data),
              post: post,
              message: message,
              owner: owner,
              auditor: auditor,
              auditorAsNull: auditorAsNull,
              withQuery: withQuery))
          .toList();
  Comment._(int id, this.post, this.message, this.owner, this.auditor)
      : super(id);

  final int post;
  final String message;
  final int owner;
  final int? auditor;

  @override
  int compareTo(Comment other) =>
      compare(post, other.post) ??
      compare(message, other.message) ??
      compare(owner, other.owner) ??
      compare(auditor, other.auditor) ??
      0;

  @override
  Comment copyWith(
          {int? id,
          int? post,
          String? message,
          int? owner,
          int? auditor,
          bool auditorAsNull = false}) =>
      Comment._(
          id ?? this.id,
          post ?? this.post,
          message ?? this.message,
          owner ?? this.owner,
          auditorAsNull ? null : (auditor ?? this.auditor));

  @override
  Map<String, dynamic> toJson() => {
        'id': id,
        'post': post,
        'message': message,
        'owner': owner,
        if (auditor != null) 'auditor': auditor
      };
}

class CommentRef extends ModelRef<Comment, int, CommentRef>
    implements CommentKey {
  final _ModelFeature _feature;
  CommentRef._(this._feature, int id) : super(_feature, id);

  PostRef get post => PostRef._(_feature, _feature.get(this).post);
  set post(PostRef value) => update(post: value);
  String get message => _feature.get(this).message;
  set message(String value) => update(message: value);
  UserRef get owner => UserRef._(_feature, _feature.get(this).owner);
  set owner(UserRef value) => update(owner: value);
  UserRef? get auditor {
    final v = _feature.get(this).auditor;
    return (v != null) ? UserRef._(_feature, v) : null;
  }

  set auditor(UserRef? value) =>
      (value != null) ? update(auditor: value) : update(auditorAsNull: true);

  Future<CommentRef> fetch({bool force = false}) => _feature<_RestFeature>()
      .commentRestAdapter
      .fetch(CommentQuery(id: id), force)
      .then((_) => this);

  void update(
      {PostKey? post,
      String? message,
      UserKey? owner,
      UserKey? auditor,
      bool auditorAsNull = false}) {
    final s = Comment._(
        id,
        (post ?? this.post).id,
        (message ?? this.message),
        (owner ?? this.owner).id,
        auditorAsNull ? null : (auditor ?? this.auditor)?.id);
    _feature.put(s);
  }
}

class CommentQuery {
  final int? id;
  final int? post;
  final String? message;
  final int? owner;
  final int? auditor;
  final bool? auditorIsNull;

  CommentQuery(
      {this.id,
      this.post,
      this.message,
      this.owner,
      this.auditorIsNull,
      this.auditor});
  CommentQuery._(this.id, this.post, this.message, this.owner,
      this.auditorIsNull, this.auditor);

  bool get isAll =>
      (id == null) &&
      (post == null) &&
      (message == null) &&
      (owner == null) &&
      (auditorIsNull == null) &&
      (auditor == null);
  bool get isNotAll => !isAll;

  bool test(Comment m) =>
      (id == null || id == m.id) &&
      (post == null || post == m.post) &&
      (message == null || message == m.message) &&
      (owner == null || owner == m.owner) &&
      ((auditorIsNull == true)
          ? (m.auditor == null)
          : (auditorIsNull == false)
              ? (auditor == null)
                  ? (m.auditor != null)
                  : (auditor == m.auditor)
              : ((auditor == null) || (auditor == m.auditor)));

  @override
  bool operator ==(Object? other) =>
      identical(this, other) ||
      (other is Comment &&
          id == other.id &&
          post == other.post &&
          message == other.message &&
          owner == other.owner &&
          auditor == other.auditor);

  @override
  int get hashCode => Object.hash(Comment, id, post, message, owner, auditor);
}

class Accounts extends ModelCollection<Account, ObjectId, AccountRef> {
  final _ModelFeature _feature;
  Accounts(this._feature)
      : super(_feature, (s) => AccountRef._(_feature, s.id));
  AccountRef? operator [](ObjectId id) => getOrNull(id);
  AccountRef get(ObjectId id) =>
      getOrNull(id) ?? (throw Exception('Account($id) does not exist'));
  AccountRef? getOrNull(ObjectId id) {
    final ref = AccountRef._(_feature, id);
    return _feature.contains(ref) ? ref : null;
  }

  Iterable<AccountRef> find(
          {ObjectId? id,
          String? name,
          bool? settingsIsNull,
          String? settings}) =>
      _feature
          .getAll(AccountQuery._(id, name, settingsIsNull, settings).test)
          .map((s) => AccountRef._(_feature, s.id));

  AccountRef put({ObjectId? id, required String name, String? settings}) =>
      load(Account._(id ?? ObjectId(), name, settings));
}

class Images extends ModelCollection<Image, int, ImageRef> {
  final _ModelFeature _feature;
  Images(this._feature) : super(_feature, (s) => ImageRef._(_feature, s.id));
  ImageRef? operator [](int id) => getOrNull(id);
  ImageRef get(int id) =>
      getOrNull(id) ?? (throw Exception('Image($id) does not exist'));
  ImageRef? getOrNull(int id) {
    final ref = ImageRef._(_feature, id);
    return _feature.contains(ref) ? ref : null;
  }

  Future<void> fetch(
          {int? id,
          bool? nameIsNull,
          String? name,
          int? height,
          int? width,
          bool force = false}) =>
      _feature<_RestFeature>()
          .imageRestAdapter
          .fetch(ImageQuery._(id, nameIsNull, name, height, width), force);

  Iterable<ImageRef> find(
          {int? id, bool? nameIsNull, String? name, int? height, int? width}) =>
      _feature
          .getAll(ImageQuery._(id, nameIsNull, name, height, width).test)
          .map((s) => ImageRef._(_feature, s.id));

  ImageRef put(
          {required int id,
          String? name,
          required int height,
          required int width}) =>
      load(Image._(id, name, height, width));
}

class Users extends ModelCollection<User, int, UserRef> {
  final _ModelFeature _feature;
  Users(this._feature) : super(_feature, (s) => UserRef._(_feature, s.id));
  UserRef? operator [](int id) => getOrNull(id);
  UserRef get(int id) =>
      getOrNull(id) ?? (throw Exception('User($id) does not exist'));
  UserRef? getOrNull(int id) {
    final ref = UserRef._(_feature, id);
    return _feature.contains(ref) ? ref : null;
  }

  Future<void> fetch(
          {int? id,
          String? name,
          bool? descIsNull,
          String? desc,
          Address? home,
          bool? workIsNull,
          Address? work,
          bool force = false}) =>
      _feature<_RestFeature>().userRestAdapter.fetch(
          UserQuery._(id, name, descIsNull, desc, home, workIsNull, work),
          force);

  Iterable<UserRef> find(
          {int? id,
          String? name,
          bool? descIsNull,
          String? desc,
          Address? home,
          bool? workIsNull,
          Address? work}) =>
      _feature
          .getAll(
              UserQuery._(id, name, descIsNull, desc, home, workIsNull, work)
                  .test)
          .map((s) => UserRef._(_feature, s.id));

  UserRef put(
          {required int id,
          required String name,
          String? desc,
          required Address home,
          Address? work}) =>
      load(User._(id, name, desc, home, work));
}

class UserPosts extends ModelCollection<Post, int, PostRef> {
  final _ModelFeature _feature;
  final UserRef _user;
  UserPosts(this._feature, this._user)
      : super(_feature, (s) => PostRef._(_feature, s.id),
            (s) => s.user == _user.id);
  PostRef? operator [](int id) => getOrNull(id);
  PostRef get(int id) =>
      getOrNull(id) ?? (throw Exception('Post($id) does not exist'));
  PostRef? getOrNull(int id) {
    final ref = PostRef._(_feature, id);
    return _feature.contains(ref) ? ref : null;
  }

  Future<void> fetch({int? id, String? title, bool force = false}) =>
      _feature<_RestFeature>()
          .postRestAdapter
          .fetch(PostQuery._(id, title, _user.id), force);

  Iterable<PostRef> find({int? id, String? title}) => _feature
      .getAll(PostQuery._(id, title, _user.id).test)
      .map((s) => PostRef._(_feature, s.id));

  PostRef put({required int id, required String title}) =>
      load(Post._(id, title, _user.id));
}

class Posts extends ModelCollection<Post, int, PostRef> {
  final _ModelFeature _feature;
  Posts(this._feature) : super(_feature, (s) => PostRef._(_feature, s.id));
  PostRef? operator [](int id) => getOrNull(id);
  PostRef get(int id) =>
      getOrNull(id) ?? (throw Exception('Post($id) does not exist'));
  PostRef? getOrNull(int id) {
    final ref = PostRef._(_feature, id);
    return _feature.contains(ref) ? ref : null;
  }

  Future<void> fetch({int? id, String? title, int? user, bool force = false}) =>
      _feature<_RestFeature>()
          .postRestAdapter
          .fetch(PostQuery._(id, title, user), force);

  Iterable<PostRef> find({int? id, String? title, int? user}) => _feature
      .getAll(PostQuery._(id, title, user).test)
      .map((s) => PostRef._(_feature, s.id));

  PostRef put({required int id, required String title, required int user}) =>
      load(Post._(id, title, user));
}

class PostComments extends ModelCollection<Comment, int, CommentRef> {
  final _ModelFeature _feature;
  final PostRef _post;
  PostComments(this._feature, this._post)
      : super(_feature, (s) => CommentRef._(_feature, s.id),
            (s) => s.post == _post.id);
  CommentRef? operator [](int id) => getOrNull(id);
  CommentRef get(int id) =>
      getOrNull(id) ?? (throw Exception('Comment($id) does not exist'));
  CommentRef? getOrNull(int id) {
    final ref = CommentRef._(_feature, id);
    return _feature.contains(ref) ? ref : null;
  }

  Future<void> fetch(
          {int? id,
          String? message,
          int? owner,
          bool? auditorIsNull,
          int? auditor,
          bool force = false}) =>
      _feature<_RestFeature>().commentRestAdapter.fetch(
          CommentQuery._(id, _post.id, message, owner, auditorIsNull, auditor),
          force);

  Iterable<CommentRef> find(
          {int? id,
          String? message,
          int? owner,
          bool? auditorIsNull,
          int? auditor}) =>
      _feature
          .getAll(CommentQuery._(
                  id, _post.id, message, owner, auditorIsNull, auditor)
              .test)
          .map((s) => CommentRef._(_feature, s.id));

  CommentRef put(
          {required int id,
          required String message,
          required int owner,
          int? auditor}) =>
      load(Comment._(id, _post.id, message, owner, auditor));
}

class Comments extends ModelCollection<Comment, int, CommentRef> {
  final _ModelFeature _feature;
  Comments(this._feature)
      : super(_feature, (s) => CommentRef._(_feature, s.id));
  CommentRef? operator [](int id) => getOrNull(id);
  CommentRef get(int id) =>
      getOrNull(id) ?? (throw Exception('Comment($id) does not exist'));
  CommentRef? getOrNull(int id) {
    final ref = CommentRef._(_feature, id);
    return _feature.contains(ref) ? ref : null;
  }

  Future<void> fetch(
          {int? id,
          int? post,
          String? message,
          int? owner,
          bool? auditorIsNull,
          int? auditor,
          bool force = false}) =>
      _feature<_RestFeature>().commentRestAdapter.fetch(
          CommentQuery._(id, post, message, owner, auditorIsNull, auditor),
          force);

  Iterable<CommentRef> find(
          {int? id,
          int? post,
          String? message,
          int? owner,
          bool? auditorIsNull,
          int? auditor}) =>
      _feature
          .getAll(
              CommentQuery._(id, post, message, owner, auditorIsNull, auditor)
                  .test)
          .map((s) => CommentRef._(_feature, s.id));

  CommentRef put(
          {required int id,
          required int post,
          required String message,
          required int owner,
          int? auditor}) =>
      load(Comment._(id, post, message, owner, auditor));
}

class _ModelFeature extends WatchableStateFeature {
  late final settings = () {
    if (!contains(const SettingsKey())) {
      put(Settings._('initial settings name', null, null));
    }
    return SettingsRef._(this);
  }();
  late final config = ConfigRef._(this);
  late final accounts = Accounts(this);
  late final images = Images(this);
  late final users = Users(this);
  late final posts = Posts(this);
  late final comments = Comments(this);
  void load(data) {
    if (data is Settings) {
      put(data);
    } else if (data is Config) {
      put(data);
    } else if (data is Account) {
      put(data);
    } else if (data is Image) {
      put(data);
    } else if (data is User) {
      put(data);
    } else if (data is Post) {
      put(data);
    } else if (data is Comment) {
      put(data);
    } else if (data is Iterable) {
      data.forEach((e) => load(e));
    } else if (data is Map<String, dynamic>) {
      for (final key in data.keys) {
        if (key == 'settings' && data[key] is Map<String, dynamic>) {
          load(Settings.fromJson(data['settings']));
        } else if (key == 'config' && data[key] is Map<String, dynamic>) {
          load(Config.fromJson(data['config']));
        } else if (key == 'account' && data[key] is Map<String, dynamic>) {
          load(Account.fromJson(data['account']));
        } else if (key == 'image' && data[key] is Map<String, dynamic>) {
          load(Image.fromJson(data['image']));
        } else if (key == 'user' && data[key] is Map<String, dynamic>) {
          load(User.fromJson(data['user']));
        } else if (key == 'post' && data[key] is Map<String, dynamic>) {
          load(Post.fromJson(data['post']));
        } else if (key == 'comment' && data[key] is Map<String, dynamic>) {
          load(Comment.fromJson(data['comment']));
        } else if (key == 'accounts' && data[key] is Iterable) {
          load(Account.fromJsonList(data['accounts']));
        } else if (key == 'images' && data[key] is Iterable) {
          load(Image.fromJsonList(data['images']));
        } else if (key == 'users' && data[key] is Iterable) {
          load(User.fromJsonList(data['users']));
        } else if (key == 'posts' && data[key] is Iterable) {
          load(Post.fromJsonList(data['posts']));
        } else if (key == 'comments' && data[key] is Iterable) {
          load(Comment.fromJsonList(data['comments']));
        } else {
          throw Exception(
              'invalid map entry during load: ($key, ${data[key]})');
        }
      }
    } else {
      throw Exception('attempted to load unknown type: $data');
    }
  }

  void unload(data) {
    if (data is SettingsKey) {
      remove(data);
    } else if (data is ConfigKey) {
      remove(data);
    } else if (data is AccountKey) {
      remove(data);
    } else if (data is ImageKey) {
      remove(data);
    } else if (data is UserKey) {
      remove(data);
    } else if (data is PostKey) {
      remove(data);
    } else if (data is CommentKey) {
      remove(data);
    } else if (data is Iterable) {
      data.forEach((e) => unload(e));
    } else if (data is Map<String, dynamic>) {
      for (final key in data.keys) {
        if (key == 'settings' && data[key] is String) {
          unload(SettingsKey);
        } else if (key == 'config' && data[key] is String) {
          unload(ConfigKey);
        } else if (key == 'account' && data[key] is String) {
          unload(AccountKey.parse(data[key]));
        } else if (key == 'image' && data[key] is int) {
          unload(ImageKey(data[key] as int));
        } else if (key == 'user' && data[key] is int) {
          unload(UserKey(data[key] as int));
        } else if (key == 'post' && data[key] is int) {
          unload(PostKey(data[key] as int));
        } else if (key == 'comment' && data[key] is int) {
          unload(CommentKey(data[key] as int));
        } else if (key == 'accounts' && data[key] is Iterable<String>) {
          (data[key] as Iterable)
              .forEach((e) => unload(AccountKey.parse(data[key])));
        } else if (key == 'images' && data[key] is Iterable<int>) {
          (data[key] as Iterable)
              .forEach((e) => unload(ImageKey(data[key] as int)));
        } else if (key == 'users' && data[key] is Iterable<int>) {
          (data[key] as Iterable)
              .forEach((e) => unload(UserKey(data[key] as int)));
        } else if (key == 'posts' && data[key] is Iterable<int>) {
          (data[key] as Iterable)
              .forEach((e) => unload(PostKey(data[key] as int)));
        } else if (key == 'comments' && data[key] is Iterable<int>) {
          (data[key] as Iterable)
              .forEach((e) => unload(CommentKey(data[key] as int)));
        } else {
          throw Exception(
              'invalid map entry during unload: ($key, ${data[key]})');
        }
      }
    } else {
      throw Exception('attempted to unload unknown type: $data');
    }
  }

  @override
  void onEvents(StateEvents events) {
    for (final v in events.values) {
      if (v is Settings) {
        if (v.account != null) {
          _onRipple(AccountKey(v.account!));
        }
        if (v.user != null) {
          _onRipple(UserKey(v.user!));
        }
      } else if (v is Account) {
        if (v.settings != null) {
          _onRipple(SettingsKey());
        }
      } else if (v is Post) {
        _onRipple(UserKey(v.user));
      } else if (v is Comment) {
        _onRipple(PostKey(v.post));
        _onRipple(UserKey(v.owner));
        if (v.auditor != null) {
          _onRipple(UserKey(v.auditor!));
        }
      }
    }
    super.onEvents(events);
  }

  void _onRipple(StateKey k) {
    if (addRipple(k)) {
      if (k is SettingsKey) {
        final settings = getOrNull(k);
        final settings_account = settings?.account;
        if (settings_account != null) {
          _onRipple(AccountKey(settings_account));
        }
        final settings_user = settings?.user;
        if (settings_user != null) {
          _onRipple(UserKey(settings_user));
        }
      } else if (k is AccountKey) {
        final account = getOrNull(k);
        final account_settings = account?.settings;
        if (account_settings != null) {
          _onRipple(SettingsKey());
        }
      } else if (k is PostKey) {
        final post = getOrNull(k);
        final post_user = post?.user;
        if (post_user != null) {
          _onRipple(UserKey(post_user));
        }
      } else if (k is CommentKey) {
        final comment = getOrNull(k);
        final comment_post = comment?.post;
        if (comment_post != null) {
          _onRipple(PostKey(comment_post));
        }
        final comment_owner = comment?.owner;
        if (comment_owner != null) {
          _onRipple(UserKey(comment_owner));
        }
        final comment_auditor = comment?.auditor;
        if (comment_auditor != null) {
          _onRipple(UserKey(comment_auditor));
        }
      }
    }
  }
}

class _TriggerFeature extends StateFeature {
  @override
  void onEvent(StateEvent e) {
    if (e.isRemove) {
      _onRemove(e);
    }
  }

  void _onRemove(StateEvent e) {
    if (e is StateEvent<User>) {
      _log.trace(() => 'removed-User');
      for (final post in getAll<Post>((s) => s.user == e.value.id)) {
        e.source.remove(post);
      }
      for (final comment in getAll<Comment>((s) => s.owner == e.value.id)) {
        e.source.remove(comment);
      }
      for (final comment in getAll<Comment>((s) => s.auditor == e.value.id)) {
        e.source.remove(comment);
      }
    } else if (e is StateEvent<Post>) {
      _log.trace(() => 'removed-Post');
      for (final comment in getAll<Comment>((s) => s.post == e.value.id)) {
        e.source.remove(comment);
      }
    }
  }
}
